from __future__ import annotations

from functools import reduce
from typing import Any, Callable, Protocol, TypeVar, overload

from .core import Pattern

T = TypeVar("T")
T1 = TypeVar("T1")


def Index(
    pat: Pattern[list[T]],
    index: int,
) -> Pattern[T]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return _match(input_)[index]

    _new.match = match.__get__(_new)
    _new.alias = f"{_new}[{index}]"

    return _new  # type: ignore


def Slice(
    pat: Pattern[list[T]],
    start: int | None = None,
    end: int | None = None,
    step: int = 1,
) -> Pattern[list[T]]:
    if start is None and end is None:
        return pat
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return _match(input_)[start:end:step]

    _new.match = match.__get__(_new)
    if start is not None and end is not None:
        _slice = f"{start}:{end}"
    elif start is not None:
        _slice = f"{start}:"
    else:
        _slice = f":{end}"
    if step != 1:
        _slice += f":{step}"
    _new.alias = f"{_new}[{_slice}]"

    return _new


def Map(
    pat: Pattern[list[T]],
    func: Callable[[T], T1],
    funcname: str | None = None,
) -> Pattern[list[T1]]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return list(map(func, _match(input_)))

    _new.match = match.__get__(_new)
    _new.alias = f"{_new}.map({funcname or func.__name__})"

    return _new  # type: ignore


def Filter(
    pat: Pattern[list[T]],
    func: Callable[[T], bool],
    funcname: str | None = None,
) -> Pattern[list[T]]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return list(filter(func, _match(input_)))

    _new.match = match.__get__(_new)
    _new.alias = f"{_new}.filter({funcname or func.__name__})"

    return _new


_T_contra = TypeVar("_T_contra", contravariant=True)
_T_co = TypeVar("_T_co", covariant=True)


class SupportsAdd(Protocol[_T_contra, _T_co]):
    def __add__(self, x: _T_contra, /) -> _T_co: ...


class SupportsRAdd(Protocol[_T_contra, _T_co]):
    def __radd__(self, x: _T_contra, /) -> _T_co: ...


class _SupportsSumWithNoDefaultGiven(SupportsAdd[Any, Any], SupportsRAdd[int, Any], Protocol): ...


_SupportsSumNoDefaultT = TypeVar("_SupportsSumNoDefaultT", bound=_SupportsSumWithNoDefaultGiven)


def Sum(pat: Pattern[list[_SupportsSumNoDefaultT]]) -> Pattern[_SupportsSumNoDefaultT]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return sum(_match(input_))

    _new.match = match.__get__(_new)
    _new.alias = f"sum({_new})"

    return _new  # type: ignore


@overload
def Reduce(
    pat: Pattern[list[T]],
    func: Callable[[T, T], T],
    initializer: None = ...,
    funcname: str | None = ...,
) -> Pattern[T]: ...


@overload
def Reduce(
    pat: Pattern[list[T]],
    func: Callable[[T1, T], T1],
    initializer: T1,
    funcname: str | None = ...,
) -> Pattern[T1]: ...


def Reduce(
    pat: Pattern[list[T]],
    func: Callable[[T, T], T] | Callable[[T1, T], T1],
    initializer: T1 | None = None,
    funcname: str | None = None,
) -> Pattern:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return reduce(func, _match(input_), initializer) if initializer is not None else reduce(func, _match(input_))  # type: ignore

    _new.match = match.__get__(_new)
    _new.alias = f"{_new}.reduce({funcname or func.__name__})"

    return _new  # type: ignore


def Join(
    pat: Pattern[list[str]],
    sep: str,
) -> Pattern[str]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return sep.join(_match(input_))

    _new.match = match.__get__(_new)
    _new.alias = f"{_new}.join({sep!r})"

    return _new  # type: ignore


def Upper(
    pat: Pattern[str],
) -> Pattern[str]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return _match(input_).upper()

    _new.match = match.__get__(_new)
    _new.alias = f"{_new}.upper()"

    return _new  # type: ignore


def Lower(
    pat: Pattern[str],
) -> Pattern[str]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return _match(input_).lower()

    _new.match = match.__get__(_new)
    _new.alias = f"{_new}.lower()"

    return _new  # type: ignore


def Dot(
    pat: Pattern[Any],
    origin: type[T],
    key: str,
    default: T | None = None,
) -> Pattern[T]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return getattr(_match(input_), key, default)

    _new.match = match.__get__(_new)
    _new.alias = f"{_new}.{key}"

    _new.origin = origin
    return _new  # type: ignore


def GetItem(
    pat: Pattern[Any],
    origin: type[T],
    key: str,
    default: T | None = None,
) -> Pattern[T]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_) -> T:
        try:
            return _match(input_)[key]
        except Exception as e:  # pragma: no cover
            if default is not None:
                return default
            raise e

    _new.match = match.__get__(_new)
    _new.alias = f"{_new}.{key}"

    _new.origin = origin
    return _new  # type: ignore


def Step(
    pat: Pattern[T],
    func: Callable[[T], T1],
    *args,
    funcname: str | None = None,
    **kwargs,
) -> Pattern[T1]:
    _new = pat.copy()
    _match = _new.match

    def match(self, input_):
        return func(_match(input_), *args, **kwargs)

    _new.match = match.__get__(_new)
    _new.alias = f"{funcname or func.__name__}({_new})"

    return _new  # type: ignore



from __future__ import annotations

from collections.abc import Mapping as ABCMap
from collections.abc import MutableMapping as ABCMuMap
from collections.abc import MutableSequence as ABCMuSeq
from collections.abc import MutableSet as ABCMuSet
from collections.abc import Sequence as ABCSeq
from collections.abc import Set as ABCSet
from contextlib import suppress
from copy import deepcopy
import inspect
from types import FunctionType, LambdaType, MethodType
from typing import Any, ForwardRef, Literal, Protocol, TypeVar, Union, overload, runtime_checkable
from typing_extensions import Annotated, TypeForm, get_args, get_origin

from tarina.lang import lang

from .base import (
    ANY,
    NONE,
    DirectPattern,
    DirectTypePattern,
    ForwardRefPattern,
    RegexPattern,
    SwitchPattern,
    UnionPattern,
    combine,
)
from .context import all_patterns
from .core import Pattern
from .util import CGenericAlias, CUnionType, GenericAlias, RawStr, TPattern

_Contents = (Union, CUnionType, Literal)


def _generic_parser(item: GenericAlias, extra: str) -> Pattern:  # type: ignore
    origin = get_origin(item)
    if origin is Annotated:
        org, *meta = get_args(item)
        if switch := next((i for i in meta if isinstance(i, dict)), None):
            return SwitchPattern(switch)
        if not isinstance(_o := parser(org, extra), Pattern):  # type: ignore  # pragma: no cover
            raise TypeError(_o)
        validators = [i for i in meta if callable(i)]  # pragma: no cover
        return combine(
            _o,
            alias=al[-1] if (al := [i for i in meta if isinstance(i, str)]) else _o.alias,
            validator=(lambda x: all(i(x) for i in validators)) if validators else None,
        )
    if origin in _Contents:
        _args = {parser(t, extra) for t in get_args(item)}  # pragma: no cover
        return (_args.pop() if len(_args) == 1 else UnionPattern(*_args)) if _args else ANY
    if origin in (list, tuple, set, dict, type, frozenset):
        item = origin[get_args(item)]  # type: ignore
    return Pattern(origin=item, alias=f"{repr(item).split('.')[-1]}").accept(item)


def _typevar_parser(item: TypeVar):
    return Pattern(alias=f"{item}"[1:]).accept(item)


def _protocol_parser(item: type):
    if not getattr(item, "_is_runtime_protocol", True):  # pragma: no cover
        item = runtime_checkable(deepcopy(item))  # type: ignore
    return Pattern(alias=f"{item}").accept(item)


_T = TypeVar("_T")
_K = TypeVar("_K")


@overload
def parser(item: TPattern, extra: str = "allow") -> RegexPattern: ...


@overload
def parser(item: list[_T] | tuple[_T, ...] | set[_T], extra: str = "allow") -> UnionPattern[_T]: ...


@overload
def parser(item: dict[_K, _T], extra: str = "allow") -> SwitchPattern[_T, _K]: ...


@overload
def parser(item: TypeVar, extra: str = "allow") -> Pattern[Any]: ...


@overload
def parser(item: type[Protocol], extra: str = "allow") -> Pattern[Any]: ...  # type: ignore


@overload
def parser(item: FunctionType | MethodType | LambdaType, extra: str = "allow") -> Pattern: ...


@overload
def parser(item: TypeForm[_T], extra: str = "allow") -> Pattern[_T]: ...


# @overload
# def parser(item: CUnionType, extra: str = "allow") -> UnionPattern[Any]: ...


@overload
def parser(item: _T, extra: str = "allow") -> Pattern[_T]: ...


def parser(item: Any, extra: str = "allow") -> Pattern:
    """将一般数据类型转为 Pattern 或者特殊类型"""
    if isinstance(item, Pattern):
        return item
    with suppress(TypeError):
        if item and (pat := all_patterns().get(item, None)):
            return pat
    if isinstance(item, (GenericAlias, CGenericAlias, CUnionType)):
        return _generic_parser(item, extra)
    if isinstance(item, TypeVar):
        return _typevar_parser(item)
    if getattr(item, "_is_protocol", False):
        return _protocol_parser(item)
    if isinstance(item, (FunctionType, MethodType, LambdaType)):
        if len((sig := inspect.signature(item)).parameters) not in (1, 2):  # pragma: no cover
            raise TypeError(f"{item} can only accept 1 or 2 argument")
        anno = list(sig.parameters.values())[-1].annotation
        return (
            Pattern((Any if sig.return_annotation == inspect.Signature.empty else sig.return_annotation))  # type: ignore
            .accept(Any if anno == inspect.Signature.empty else anno)
            .convert(item if len(sig.parameters) == 2 else lambda _, x: item(x))
        )
    if isinstance(item, TPattern):  # type: ignore
        return RegexPattern(item.pattern, alias=f"'{item.pattern}'")
    if isinstance(item, str):
        if item.startswith("re:"):
            pat = item[3:]
            return Pattern.regex_match(pat, alias=f"'{pat}'")
        if item.startswith("rep:"):
            pat = item[4:]
            return RegexPattern(pat, alias=f"'{pat}'")
        if "|" in item:
            names = item.split("|")
            return UnionPattern(*(all_patterns().get(i, i) for i in names if i))
        return DirectPattern(item, alias=f"'{item}'")
    if isinstance(item, RawStr):
        return DirectPattern(item.value, alias=f"'{item.value}'")
    if isinstance(item, (list, tuple, set, ABCSeq, ABCMuSeq, ABCSet, ABCMuSet)):  # Args[foo, [123, int]]
        return UnionPattern(*map(lambda x: parser(x) if inspect.isclass(x) else x, item))
    if isinstance(item, (dict, ABCMap, ABCMuMap)):
        return SwitchPattern(dict(item))
    if isinstance(item, ForwardRef):
        return ForwardRefPattern(item)
    if item is None or type(None) == item:
        return NONE
    if extra == "ignore":
        return ANY
    elif extra == "reject":
        raise TypeError(lang.require("nepattern", "parse_reject").format(target=item))
    if inspect.isclass(item):
        return DirectTypePattern(origin=item)  # type: ignore
    return DirectPattern(item)


__all__ = ["parser"]



from __future__ import annotations

from copy import deepcopy
import re
from typing import Any, Callable, Generic, TypeVar, Union, cast, overload
from typing_extensions import Self, TypeForm

from tarina import Empty, generic_isinstance
from tarina.lang import lang

from .exception import MatchFailed
from .util import TPattern

T = TypeVar("T")
_T = TypeVar("_T")


class ValidateResult(Generic[T]):
    """参数表达式验证结果"""

    def __init__(
        self,
        value: T | type[Empty] = Empty,
        error: Exception | type[Empty] = Empty,
    ):
        self._value = value
        self._error = error

    __slots__ = ("_value", "_error")

    def value(self) -> T:
        """获取验证结果"""
        if self._value is Empty:
            raise RuntimeError("cannot access value")
        return self._value  # type: ignore

    def error(self) -> Exception | None:
        """获取验证错误"""
        if self._error is not Empty:
            assert isinstance(self._error, Exception)
            return self._error

    @property
    def success(self) -> bool:
        """是否验证成功"""
        return self._value is not Empty

    @property
    def failed(self) -> bool:
        """是否验证失败"""
        return self._value is Empty

    def __bool__(self):  # pragma: no cover
        return self._value != Empty

    def __repr__(self):
        return (
            f"ValidateResult(value={self._value!r})"
            if self._value is not Empty
            else f"ValidateResult(error={self._error!r})"
        )


class Pattern(Generic[T]):
    @staticmethod
    def regex_match(pattern: str | TPattern, alias: str | None = None) -> _RegexPattern[str]:
        """构建一个仅正则表达式匹配的 Pattern，不进行转换"""
        pat = _RegexPattern(pattern, str, alias or str(pattern))

        @pat.convert
        def _(self: _RegexPattern, x: str):
            mat = re.match(self.pattern, x) or re.search(self.pattern, x)
            if not mat:
                raise MatchFailed(
                    lang.require("nepattern", "error.content").format(target=x, expected=self.pattern)
                )
            return mat[0]

        return pat

    @staticmethod
    def regex_convert(
        pattern: str,
        origin: TypeForm[_T],
        fn: Callable[[re.Match[str]], _T],
        alias: str | None = None,
        allow_origin: bool = False,
    ) -> _RegexPattern[_T]:
        """构建一个正则表达式匹配的 Pattern，并提供转换函数"""
        pat = _RegexPattern(pattern, origin, alias or str(pattern))
        if allow_origin:
            pat.accept(Union[str, origin])

            @pat.convert
            def _(self: _RegexPattern, x):
                if isinstance(x, cast(type[_T], origin)):
                    return x
                mat = re.match(self.pattern, x) or re.search(self.pattern, x)
                if not mat:
                    raise MatchFailed(
                        lang.require("nepattern", "error.content").format(target=x, expected=self.pattern)
                    )
                return fn(mat)

        else:
            pat.accept(str)

            @pat.convert
            def _(self: _RegexPattern, x: str):
                mat = re.match(self.pattern, x) or re.search(self.pattern, x)
                if not mat:
                    raise MatchFailed(
                        lang.require("nepattern", "error.content").format(target=x, expected=self.pattern)
                    )
                return fn(mat)

        return pat

    @staticmethod
    def on(obj: _T) -> Pattern[_T]:
        """提供 DataUnit 类型的构造方法"""
        from .base import DirectPattern

        return DirectPattern(obj)

    @overload
    def __init__(self, origin: TypeForm[T], alias: str | None = None): ...

    @overload
    def __init__(self: Pattern[Any], *, alias: str | None = None): ...

    def __init__(self, origin: Any | None = None, alias: str | None = None):
        self.origin: type[T] = origin or Any  # type: ignore
        self.alias = alias

        self._accepts = Any
        self._post_validator = None
        self._pre_validator = (lambda x: generic_isinstance(x, self.origin)) if origin else None
        self._converter = None
        self._pre_validate_modified = False

    def __init_subclass__(cls, **kwargs):
        cls.__hash__ = Pattern.__hash__

    def accept(self, input_type: Any):
        """设置接受的输入类型"""
        if input_type is ...:
            input_type = Any
        self._accepts = input_type
        if not self._pre_validate_modified:
            self._pre_validator = None
        return self

    def pre_validate(self, func: Callable[[Any], bool]):
        """设置预验证函数 (经过 accept 后，convert 前)"""
        self._pre_validator = func
        self._pre_validate_modified = True
        return self

    def post_validate(self, func: Callable[[T], bool]):
        """设置后验证函数 (convert 后，仅当设置了 converter 才会生效)"""
        self._post_validator = func
        return self

    def convert(self, func: Callable[[Self, Any], T | None]):
        """设置转换函数, 返回 None 时表示转换失败"""
        self._converter = func
        return self

    def match(self, input_: Any) -> T:
        if not generic_isinstance(input_, self._accepts):
            raise MatchFailed(
                lang.require("nepattern", "error.type").format(target=input_, expected=self._accepts)
            )
        if self._pre_validator and not self._pre_validator(input_):
            raise MatchFailed(
                lang.require("nepattern", "error.content").format(target=input_, expected=self.origin)
            )
        if self._converter:
            input_ = self._converter(self, input_)
            if input_ is None:
                raise MatchFailed(
                    lang.require("nepattern", "error.content").format(target=input_, expected=self.origin)
                )
            if self._post_validator and not self._post_validator(input_):
                raise MatchFailed(
                    lang.require("nepattern", "error.content").format(target=input_, expected=self.origin)
                )
        return input_

    def execute(self, input_: Any) -> ValidateResult[T]:
        """执行验证"""
        try:
            return ValidateResult(self.match(input_))
        except Exception as e:
            return ValidateResult(error=e)

    def __str__(self):
        if self.alias:
            return self.alias
        if self._accepts is self.origin or self._accepts is Any:
            return f"{getattr(self.origin, '__name__', self.origin)}"
        return f"{getattr(self._accepts, '__name__', self._accepts)} -> {getattr(self.origin, '__name__', self.origin)}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.origin}, {self.alias!r})"

    def copy(self) -> Self:
        return deepcopy(self)

    def __lshift__(self, other):  # pragma: no cover
        return self.execute(other)

    def __rmatmul__(self, other) -> Self:  # pragma: no cover
        if isinstance(other, str):
            self.alias = other
        return self

    def __matmul__(self, other) -> Self:  # pragma: no cover
        if isinstance(other, str):
            self.alias = other
        return self

    def __hash__(self):
        return hash((self.origin, self.alias, self._accepts, self._converter))

    def __eq__(self, other):
        return isinstance(other, Pattern) and self.__hash__() == other.__hash__()


class _RegexPattern(Pattern[T]):
    def __init__(self, pattern: str | TPattern, origin: TypeForm[T], alias: str | None = None):
        super().__init__(origin, alias)
        _pat = pattern if isinstance(pattern, str) else pattern.pattern
        if _pat.startswith("^") or _pat.endswith("$"):
            raise ValueError(lang.require("nepattern", "error.pattern_head_or_tail").format(target=pattern))
        if isinstance(pattern, str):
            self.pattern = f"^{pattern}$"
        else:
            self.pattern = re.compile(f"^{pattern.pattern}$", pattern.flags)

    def prefixed(self):
        """转为前缀型匹配"""
        new = self.copy()
        if isinstance(self.pattern, str):
            new.pattern = self.pattern[:-1]
        else:  # pragma: no cover
            new.pattern = re.compile(self.pattern.pattern[:-1], self.pattern.flags)
        return new

    def suffixed(self):
        """转为后缀型匹配"""
        new = self.copy()
        if isinstance(self.pattern, str):
            new.pattern = self.pattern[1:]
        else:  # pragma: no cover
            new.pattern = re.compile(self.pattern.pattern[1:], self.pattern.flags)
        return new
